﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace DWZB_Agv_Agreement.Graph
{
    /// <summary>
    /// 广度优先遍历与最短路径
    /// </summary>
    public class ShortestPath
    {
        /// <summary>
        /// 图的引用
        /// </summary>
        private readonly IGraph G;
        /// <summary>
        /// 起始点
        /// </summary>
        private readonly int s;
        /// <summary>
        /// 记录dfs的过程中节点是否被访问
        /// </summary>
        private readonly bool[] visited;
        /// <summary>
        /// 记录路径, from[i]表示查找的路径上i的上一个节点
        /// </summary>
        private readonly int[] from;
        /// <summary>
        /// 记录路径中节点的次序。ord[i]表示i节点在路径中的次序。
        /// </summary>
        private readonly int[] ord;

        /// <summary>
        /// 构造函数, 寻路算法, 寻找图graph从s点到其他点的路径
        /// </summary>
        /// <param name="graph">构建的抽象图实例</param>
        /// <param name="s">起点</param>
        public ShortestPath(IGraph graph, int s)
        {
            // 算法初始化
            G = graph;
            if (s < 0 || s >= G.V())
            {
                throw new Exception("数据校验错误！");
            }

            visited = new bool[G.V()];
            from = new int[G.V()];
            ord = new int[G.V()];
            for (int i = 0; i < G.V(); i++)
            {
                visited[i] = false;
                from[i] = -1;
                ord[i] = -1;
            }
            this.s = s;
            // 无向图最短路径算法, 从s开始广度优先遍历整张图
            Queue<int> q = new();
            q.Enqueue(s);
            visited[s] = true;
            ord[s] = 0;
            while (q.Count > 0)
            {
                int v = q.Dequeue();
                foreach (int i in G.Adj(v))
                {
                    if (!visited[i])
                    {
                        q.Enqueue(i);
                        visited[i] = true;
                        from[i] = v;
                        ord[i] = ord[v] + 1;
                    }
                }
            }
        }

        /// <summary>
        /// 查询从s点到w点是否有路径
        /// </summary>
        /// <param name="w">目标点</param>
        /// <returns></returns>
        public bool HasPath(int w)
        {
            if (w < 0 || w >= G.V())
            {
                throw new Exception("数据校验错误！");
            }
            return visited[w];
        }

        /// <summary>
        /// 查询从s点到w点的路径, 存放在vec中
        /// </summary>
        /// <param name="w">目标点</param>
        /// <returns></returns>
        public List<int> Path(int w)
        {
            if (!HasPath(w))
            {
                throw new Exception("数据校验错误！");
            }
            Stack<int> s = new();
            // 通过from数组逆向查找到从s到w的路径, 存放到栈中
            int p = w;
            while (p != -1)
            {
                s.Push(p);
                p = from[p];
            }

            // 从栈中依次取出元素, 获得顺序的从s到w的路径
            List<int> res = new();
            while (s.Count > 0)
            {
                res.Add(s.Pop());
            }

            return res;
        }

        /// <summary>
        /// 打印出从s点到w点的路径
        /// </summary>
        /// <param name="w">目标点</param>
        public string ShowPath(int w)
        {
            if (!HasPath(w))
            {
                throw new Exception("数据校验错误！");
            }
            List<int> vec = Path(w);
            string pathStr = string.Empty;
            for (int i = 0; i < vec.Count; i++)
            {
                Console.WriteLine(vec[i]);
                pathStr += vec[i];
                if (i == vec.Count - 1)
                {
                    Console.WriteLine();
                }
                else
                {
                    Console.WriteLine(" -> ");
                    pathStr += " -> ";
                }
            }
            return pathStr;
        }

        /// <summary>
        ///  查看从s点到w点的最短路径长度
        ///  若从s到w不可达，返回-1
        /// </summary>
        /// <param name="w">目标点</param>
        /// <returns></returns>
        public int Length(int w)
        {
            if (w < 0 || w >= G.V())
            {
                throw new Exception("数据校验错误！");
            }
            return ord[w];
        }
    }
}
