﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace DWZB_Agv_Agreement
{
    /// <summary>
    /// RS485转Wifi通信类
    /// </summary>
    public class RS485ToWifiModbus
    {
        #region [private member]
        private readonly string m_iP;
        private readonly int m_iPort;
        private TcpClient m_tcpClient = null;
        private readonly byte[] m_RecBuffer = new byte[1024];
        private readonly object m_lock = new object();
        private readonly Queue<byte[]> m_QBuffer = new Queue<byte[]>();
        //private int m_iCount = 0;
        public byte[] QueryCmdArr { get; set; }
        #endregion [private member]

        #region [Ctor]

        public RS485ToWifiModbus(string IP, int Port)
        {
            m_iP = IP;
            m_iPort = Port;
        }

        #endregion [Ctor]

        #region [Event]
        /// <summary>
        /// 模块连接事件
        /// </summary>
        /// <param name="bState"></param>
        public event Action<bool, string> OnConnected;
        /// <summary>
        /// 数据解析事件，对收到的数据进行解析
        /// </summary>
        /// <param name="buffer">接收的数据</param>
        /// <param name="datalen">接收的数据长度</param>
        public event Action<byte[], int> OnDataParse;

        #endregion [Event]

        #region [public property]

        /// <summary>
        /// 返回模块的连接状态
        /// </summary>
        public bool IsConnected => m_tcpClient != null && m_tcpClient.Connected;

        #endregion [public property]

        #region [public method]

        /// <summary>
        /// 启动模块任务
        /// </summary>
        public void BeginWork(object sender, DoWorkEventArgs e)
        {
            BackgroundWorker thWork = sender as BackgroundWorker;
            while (true)
            {
                if (thWork.CancellationPending)
                {
                    e.Cancel = true;
                    return;
                }

                #region [确定连接模块]
                if (!IsConnected)
                {
                    Thread.Sleep(9100);
                    OnConnected?.Invoke(false, "Socket连接已经断开，开始尝试重新连接......");
                    _ = InitClient();
                    continue;
                }
                #endregion [确定连接模块]

                #region 接收数据
                try
                {
                    byte[] Cmd = GetCmd();
                    if (Cmd == null)
                    {
                        Cmd = QueryCmdArr;
                    }
                    if (Cmd != null)
                    {
                        _ = m_tcpClient.Client.Send(Cmd);
                        Thread.Sleep(800);
                        Array.Clear(m_RecBuffer, 0, m_RecBuffer.Length);
                        int iret = m_tcpClient.Client.Receive(m_RecBuffer, SocketFlags.None);

                        if (iret > 0)
                        {
                            OnDataParse?.Invoke(m_RecBuffer, iret);
                        }
                    }
                }
                catch (Exception ex)
                {
                    OnConnected?.Invoke(false, $"Socket通讯发生异常！异常信息：【{ex.Message}】【{ex.StackTrace}】");
                }
                #endregion

                Thread.Sleep(200);
            }
        }
        /// <summary>
        /// 添加一条指令,向模块缓冲区中添加一条指令
        /// </summary>
        /// <param name="array"></param>
        public void AddCmd(byte[] array)
        {
            lock (m_lock)
            {
                if (m_QBuffer.Count > 10)
                {
                    m_QBuffer.Clear();
                }
                if (array.Length == 8)
                {
                    m_QBuffer.Enqueue(array);
                }
            }
        }
        /// <summary>
        /// 获取一条指令
        /// </summary>
        /// <returns></returns>
        private byte[] GetCmd()
        {
            lock (m_lock)
            {
                if (m_QBuffer.Count > 0)
                {
                    return m_QBuffer.Dequeue();
                }
            }
            return null;
        }
        public bool SendDataBtyArr(byte[] array)
        {
            try
            {
                if (!IsConnected)
                {
                    return false;
                }
                int iretSend = m_tcpClient.Client.Send(array);
                if (iretSend == 0)
                {
                    return false;
                }
                return true;
            }
            catch
            {
                return true;
            }
        }

        #endregion [public method]
        public void CloseTcpClient()
        {
            if (m_tcpClient != null)
            {
                m_tcpClient.Close();
                m_tcpClient = null;
            }
        }
        #region [private method]
        /// <summary>
        ///   初始化模块客户端
        /// </summary>
        public bool InitClient()
        {
            try
            {
                if (m_tcpClient != null)
                {
                    m_tcpClient.Close();
                    m_tcpClient = null;
                }
                m_tcpClient = new TcpClient
                {
                    ReceiveBufferSize = 1024,
                    SendBufferSize = 1024
                };

                m_tcpClient.Connect(m_iP, m_iPort);
                if (m_tcpClient.Connected)
                {
                    OnConnected?.Invoke(true, "Socket连接成功！");
                    return true;
                }
                else
                {
                    return false;
                }
            }
            catch (Exception ex)
            {
                OnConnected?.Invoke(false, $"Socket连接异常！异常信息：【{ex.Message}】【{ex.StackTrace}】");
                return false;
            }
        }

        #endregion [private method]
    }
}
